# Copyright 2022 Ethan Kinnear
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handle tldr-page languages."""

from os import getenv
from typing import Iterable

from tldr_man.pages import TLDR_CACHE_HOME


def all_languages():
    return (
        get_language_directory(pages_dir.name.removeprefix('pages').lstrip('.') or 'en')
        for pages_dir in TLDR_CACHE_HOME.iterdir()
        if pages_dir.is_dir()
    )


def get_languages() -> Iterable[str]:
    """
    Returns a list of the user's preferred languages, inferred by the environment variables LANG and LANGUAGE.

    See https://github.com/tldr-pages/tldr/blob/main/CLIENT-SPECIFICATION.md#language for details.
    """

    languages: list[str] = []

    env_lang = getenv('LANG')
    env_language = getenv('LANGUAGE')
    env_tldr_language = getenv('TLDR_LANGUAGE')

    if env_language:
        languages += env_language.split(':')

    if env_tldr_language:
        languages += env_tldr_language.split(':')

    if env_lang and env_lang.upper() not in ['C', 'POSIX']:
        languages.insert(0, env_lang)

    languages.append('en')

    # Remove country code from the language codes
    return map(get_language_directory, languages)


def _language_code_as_parts(language_code: str) -> (str, str):
    """Removes country codes from language codes and preforms data normalization."""
    code = language_code.split('.')[0].split('_', 1)

    language = code[0].strip().lower()
    region = code[1].strip().upper() if len(code) >= 2 else language.upper()

    return language, region


def get_language_directory(language_code: str) -> str:
    """Get the name of the directory for a language code."""
    language, region = _language_code_as_parts(language_code)
    if language == 'en':
        return 'pages'
    else:
        full_locale = f'pages.{language}_{region}'
        if (TLDR_CACHE_HOME / full_locale).is_dir():
            return full_locale
        else:
            return f'pages.{language}'



# Copyright 2022 Ethan Kinnear
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interact with tldr-pages and the tldr-pages manpage cache."""

import os
import re
import zipfile
from contextlib import suppress
from pathlib import Path
from shutil import rmtree, move
from subprocess import run, PIPE, DEVNULL
from typing import Optional, Iterable

import requests
from click import secho, progressbar, style
from xdg import XDG_CACHE_HOME

from tldr_man.util import mkstemp_path, mkdtemp_path, esecho

TLDR_CACHE_DIR_NAME = 'tldr-man'

TLDR_ZIP_ARCHIVE_URL = "https://tldr.sh/assets/tldr.zip"

TLDR_MANPAGE_SECTION = '1'

MANPAGE_HEADER = """
% {name}(1) {name}
%
% tldr-man-client

# NAME
{name} - {desc}

# DESCRIPTION
{info}

# EXAMPLES
{examples}

"""[1:-1]

PANDOC_MISSING_MESSAGE = """
Error: Couldn't find the `pandoc` command.
Make sure that pandoc is installed and on your PATH.
    Installation instructions: https://pandoc.org/installing.html
"""[1:-1]

MANPAGE_MISSING_MESSAGE = """
Error: Couldn't find the `man` command.
Make sure that `man` is on your PATH.
"""

PAGE_NOT_FOUND_MESSAGE = """
Error: Page `{page_name}` is not found.
    Try running `tldr --update` to update the tldr-pages cache.

Request this page here:
    https://github.com/tldr-pages/tldr/issues/new?title=page%20request:%20{page_name}
"""[1:-1]

CACHE_DOES_NOT_EXIST_MESSAGE = """
The tldr-pages cache needs to be generated before `tldr` can be used.
    Run `tldr --update` to generate the cache.
"""[1:-1]


def tldr_cache_home(location: Path = XDG_CACHE_HOME) -> Path:
    return location / TLDR_CACHE_DIR_NAME


TLDR_CACHE_HOME: Path = tldr_cache_home()


def download_tldr_zip_archive(location: Path, url: str = TLDR_ZIP_ARCHIVE_URL) -> None:
    """Downloads the current tldr-pages zip archive into a specific location."""

    try:
        r = requests.get(url, timeout=10)
    except requests.exceptions.ConnectionError:
        esecho(f"Error: Could not make connection to {TLDR_ZIP_ARCHIVE_URL}", exitcode=1)
    except requests.exceptions.Timeout:
        esecho(f"Error: Request to {TLDR_ZIP_ARCHIVE_URL} timed out", exitcode=1)
    except requests.exceptions.RequestException:
        esecho(f"The following error occurred when trying to access {TLDR_ZIP_ARCHIVE_URL}")
        raise
    else:
        location.write_bytes(r.content)


def update_cache() -> None:
    """Updates the tldr-pages manpage cache."""

    if not pandoc_exists():
        esecho(PANDOC_MISSING_MESSAGE, exitcode=127)

    secho('Updating tldr-pages cache...', fg='cyan')

    try:
        # Create a temporary file for the tldr-pages zip archive to generate manpages from.
        tldr_zip_archive = mkstemp_path('tldr.zip')
        download_tldr_zip_archive(tldr_zip_archive)

        # Create the cache directory that will be copied to `~/.cache/tldr-man`.
        tldr_temp_dir = mkdtemp_path('tldr-man')

        # Get the zip file
        try:
            tldr_zip_path = zipfile.Path(tldr_zip_archive)
        except zipfile.BadZipFile:
            esecho(f"Error: Got a bad zipfile from {TLDR_ZIP_ARCHIVE_URL}")
            raise

        # Iterate through each language and section in the zip file.
        for language_dir in tldr_zip_path.iterdir():
            if not language_dir.is_dir():
                continue
            for sections_dir in language_dir.iterdir():

                # Get the full path to the directory where all manpages for this language and section will be extracted.
                res_dir = tldr_temp_dir / language_dir.name / sections_dir.name / ('man' + TLDR_MANPAGE_SECTION)
                res_dir.mkdir(parents=True, exist_ok=True)  # Create the directories if they don't exist.

                # Create the label for the progress bars that are shown.
                progressbar_label = (style(f'{language_dir.name:11s}', fg='blue')
                                     + ' / '
                                     + style(f'{sections_dir.name:7s}', fg='blue'))

                # Loop through each tldr-page in the current language and section (displaying a progressbar) and
                # generate a manpage for each markdown formatted tldr-page.
                with progressbar(list(sections_dir.iterdir()), label=progressbar_label) as pages:
                    for page in pages:
                        page: zipfile.Path

                        if not page.is_file():
                            continue

                        # Render and save a manpage from the tldr-page.
                        manpage = render_manpage(page.read_text())
                        (res_dir / (page.name.removesuffix('.md') + '.1')).write_text(manpage)

        # Now that the updated cache has been generated, remove the old cache, make sure the parent directory exists,
        # and move the new cache into the correct directory from the temporary directory.

        with suppress(FileNotFoundError):
            rmtree(TLDR_CACHE_HOME)

        os.makedirs(TLDR_CACHE_HOME.parent, exist_ok=True)

        move(tldr_temp_dir, TLDR_CACHE_HOME)
    finally:
        # Clean up any temporary files that aren't gone.

        with suppress(NameError, FileNotFoundError):
            # noinspection PyUnboundLocalVariable
            os.remove(tldr_zip_archive)
        with suppress(NameError, FileNotFoundError):
            # noinspection PyUnboundLocalVariable
            rmtree(tldr_temp_dir)

    secho('Done!', fg='green', bold=True)


def render_manpage(tldr_page: str) -> Optional[str]:
    """
    Render a manpage from a markdown formatted tldr-page.

    Exits with an error message if `pandoc` is not installed.
    """

    lines = tldr_page.splitlines()

    name = lines.pop(0).lstrip('# ')  # Get the name of the command.

    # Get the information of the command.
    info = []
    while (valid_line := (line := lines.pop(0)).startswith('> ')) or not info:
        if not valid_line:
            continue
        info.append(line.removeprefix('> '))

    examples = []
    for line in lines:
        if not line:
            continue

        indicator, contents = line[0], line[1:].strip().replace(r'\\', r'\e')
        match indicator:
            case '-':
                examples.append(f'\n**{contents}**\n')
            case '`':
                examples.append(f': {contents[:-1]}\n')
    examples = re.sub(r'\{\{(.*?)}}', r'*\1*', ''.join(examples))

    res = MANPAGE_HEADER.format(name=name, desc=info[0], info='\n'.join(info[1:]), examples=examples)

    try:
        return run(['pandoc', '-', '-s', '-t', 'man', '-f', 'markdown-tex_math_dollars-smart'],
                   input=res.encode('utf-8'), stdout=PIPE).stdout.decode('utf-8')
    except FileNotFoundError:
        esecho(PANDOC_MISSING_MESSAGE, exitcode=127)


def pandoc_exists() -> bool:
    """Check whether pandoc exists."""
    try:
        return run(['pandoc', '--version'], stdout=DEVNULL).returncode == 0
    except FileNotFoundError:
        return False


def verify_tldr_cache_exists():
    """Display a specific message if the tldr manpage cache doesn't exist yet, and then exit."""
    if not TLDR_CACHE_HOME.exists():
        esecho(CACHE_DOES_NOT_EXIST_MESSAGE, exitcode=1)


def display_page(page: Path) -> None:
    try:
        run(['man', page])
    except FileNotFoundError as err:
        if err.filename == 'man':
            esecho(MANPAGE_MISSING_MESSAGE, exitcode=127)
        else:
            raise


def find_page(page_name: str, /, locales: Iterable[str], page_sections: Iterable[str]) -> Optional[Path]:
    for search_dir in get_dir_search_order(locales, page_sections):
        page = search_dir / (page_name + '.1')

        if page.exists():
            return page
    else:
        esecho(PAGE_NOT_FOUND_MESSAGE.format(page_name=page_name))


def get_dir_search_order(locales: Iterable[str], page_sections: Iterable[str]) -> Iterable[Path]:
    return (
        TLDR_CACHE_HOME / locale / section / ('man' + TLDR_MANPAGE_SECTION)
        for locale in locales
        for section in page_sections
    )



#!/usr/bin/env python3

# Copyright 2022 Ethan Kinnear
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
tldr-man-client: Command-line TLDR client that displays tldr-pages as manpages.

Depends on pandoc (and man!). Install `pandoc` from https://pandoc.org/installing.html.

Run `tldr --help` for more information,
or visit the project repository at https://github.com/superatomic/tldr-man-client.
"""

from importlib import metadata
__version__ = metadata.version('tldr-man')
__author__ = "Ethan Kinnear <contact@superatomic.dev>"

import sys
from pathlib import Path
from contextlib import suppress
from os import remove
from functools import wraps
from typing import Optional

import click
from click import secho
from click_help_colors import HelpColorsCommand

from tldr_man import languages, pages
from tldr_man.util import unique, mkstemp_path, esecho


TLDR_COMMAND_NAME = 'tldr'

TLDR_PLATFORMS = 'android linux macos osx sunos windows'.split()


def click_standalone_subcommand(func):
    """Function decorator to reduce boilerplate code at the start and end of all subcommand callback functions."""
    @wraps(func)
    def wrapper(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return

        exited_with_error = False
        # noinspection PyBroadException
        try:
            func(ctx, param, value)
        except Exception:
            from traceback import format_exc
            from sys import stderr

            print(format_exc(), file=stderr)
            exited_with_error = True  # Don't call the `ctx.exit()` in the `finally` block
            ctx.exit(1)
        except SystemExit as err:
            # If `sys.exit()` was called, exit that context with the given status code.
            exited_with_error = True  # Don't call the `ctx.exit()` in the `finally` block
            ctx.exit(err.code)
        finally:
            if not exited_with_error:
                ctx.exit()

    return wrapper


@click_standalone_subcommand
def subcommand_update(_ctx, _param, _value):
    pages.update_cache()


@click_standalone_subcommand
def subcommand_render(_ctx, _param, value: str):
    page_to_render = Path(value).read_text()
    rendered_page = pages.render_manpage(page_to_render)

    try:
        path = mkstemp_path('tldr-man', text=True)
        path.write_text(rendered_page)
        pages.display_page(path)
    finally:
        with suppress(NameError, FileNotFoundError):
            # noinspection PyUnboundLocalVariable
            remove(path)


@click_standalone_subcommand
def subcommand_list(ctx, _param, _value: None):
    pages.verify_tldr_cache_exists()

    locales = get_locales(ctx)
    page_sections = get_page_sections(ctx)

    print('\n'.join(unique(
        page.stem
        for section in pages.get_dir_search_order(locales, page_sections)
        for page in section.iterdir()
        if page.is_file()
    )))


@click_standalone_subcommand
def subcommand_manpath(ctx, _param, _value: None):
    pages.verify_tldr_cache_exists()

    locales = get_locales(ctx)
    page_sections = get_page_sections(ctx)

    print(':'.join(unique(str(x.parent) for x in pages.get_dir_search_order(locales, page_sections))))


@click_standalone_subcommand
def subcommand_version(_ctx, _param, _value: str):
    secho(f'{TLDR_COMMAND_NAME} {__version__}')


@click.command(cls=HelpColorsCommand, help_headers_color='yellow', help_options_color='green')
@click.argument('page', nargs=-1, required=True)
@click.option('-p', '--platform', type=click.Choice(TLDR_PLATFORMS), is_eager=True,
              help='Override the preferred platform')
@click.option('-L', '--language', is_eager=True,
              help='Specify a preferred language')
@click.option('-u', '--update', is_flag=True, callback=subcommand_update, expose_value=False, is_eager=True,
              help='Update the tldr-pages cache')
@click.option('-r', '--render', nargs=1, callback=subcommand_render, expose_value=False, is_eager=True,
              help='Render a page locally')
@click.option('-l', '--list', is_flag=True, callback=subcommand_list, expose_value=False,
              help='List all the pages for the current platform')
@click.option('--manpath', is_flag=True, callback=subcommand_manpath, expose_value=False,
              help='Print the paths to the tldr manpages')
@click.option('-v', '-V', '--version', is_flag=True, callback=subcommand_version, expose_value=False, is_eager=True,
              help='Display the version of the client')
@click.help_option('-h', '--help')
@click.pass_context
def cli(ctx, page: list[str], **_):
    """TLDR client that displays tldr-pages as manpages"""

    pages.verify_tldr_cache_exists()

    locales = get_locales(ctx)
    page_sections = get_page_sections(ctx)

    page_name = '-'.join(page).strip().lower()

    page = pages.find_page(page_name, locales, page_sections)

    if page is not None:
        pages.display_page(page)


def get_locales(ctx) -> list[str]:
    language = ctx.params.get('language')
    if language is not None:
        page_locale = languages.get_language_directory(language)
        if page_locale not in languages.all_languages():
            esecho(f"Unrecognized locale: {language}", exitcode=1)
        else:
            return [page_locale]
    else:
        return list(languages.get_languages())


def get_page_sections(ctx) -> list[str]:
    page_sections = ['common']

    current_platform = get_current_platform()
    custom_platform = ctx.params.get('platform')

    if current_platform:
        page_sections.insert(0, current_platform)

    if custom_platform:
        page_sections.insert(0, custom_platform.replace('macos', 'osx'))

    return page_sections


def get_current_platform() -> Optional[str]:
    """Get the correct tldr platform directory name from `sys.platform`."""
    match sys.platform:
        case 'darwin':
            return 'osx'
        case 'linux':
            return 'linux'
        case 'win32' | 'cygwin' | 'msys':
            return 'windows'
        case 'sunos5':
            return 'sunos'
        case _:
            return None


if __name__ == '__main__':
    cli()



# Copyright 2022 Ethan Kinnear
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various utility functions."""

from sys import exit
from pathlib import Path
from tempfile import mkstemp, mkdtemp
from typing import Optional

from click import secho


def unique(items):
    seen = set()
    for item in items:
        if item not in seen:
            seen.add(item)
            yield item


def mkstemp_path(*args, **kwargs) -> Path:
    return Path(mkstemp(*args, **kwargs)[1])


def mkdtemp_path(*args, **kwargs) -> Path:
    return Path(mkdtemp(*args, **kwargs))


def esecho(*args, exitcode: Optional[int] = None, **kwargs):
    secho(*args, **kwargs, fg='red', err=True)
    if exitcode is not None:
        exit(exitcode)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['tldr_man']

package_data = \
{'': ['*']}

install_requires = \
['click-help-colors>=0.9.1,<0.10.0',
 'click>=8.1.3,<9.0.0',
 'requests>=2.28.1,<3.0.0',
 'xdg>=5.1.1,<6.0.0']

entry_points = \
{'console_scripts': ['tldr = tldr_man:cli', 'tldr-man = tldr_man:cli']}

setup_kwargs = {
    'name': 'tldr-man',
    'version': '1.0.2',
    'description': 'Command-line TLDR client that displays tldr-pages as manpages',
    'long_description': '<div>\n    <h1 align="center">tldr-man-client</h1>\n    <h5 align="center">A tldr-pages client that works just like <code>man</code></h5>\n</div>\n\n`tldr-man-client` is a command-line client for [tldr-pages][tldr-pages],\na collection of community-maintained help pages for command-line tools.\nIt differs from other clients because it displays its pages as `man` pages.\n\nThis client is also able to integrate with the `man` command to fall back to displaying a tldr-page for a command when\nno manpage exists.\n\nFeatures:\n- Fully abides by the [tldr-pages client specification][client-spec].\n- Supports all page languages, not just English pages.\n- Displays tldr-pages in the same style as manpages.\n- Integrates with `man` to provide a fallback for missing manpages.\n- Supports rendering markdown formatted tldr-pages with `--render`.\n- Local cache abides by the [XDG base directory specification][xdg].\n- And much more!\n\n\n## Installation\n\n### With Homebrew\n\nInstall `tldr-man-client` with [Homebrew](https://brew.sh):\n\n```shell\nbrew install superatomic/tap/tldr-man\n```\n\n### With Pip\n\nInstall `tldr-man-client` with pip (version 3.10+):\n\n```shell\npip install tldr-man\n```\n\n`tldr-man-client` additionally depends on [`pandoc`](https://pandoc.org/installing.html) being installed.\n\nAfter installation, you can view a tldr-page with the `tldr` command.\n\n\n## Usage\n\n**Display a tldr-page for a command:**\n\n```shell\ntldr <COMMAND>\n```\n\n**Update the local page cache:**\n\n```shell\ntldr --update\n```\n\n**Render a page locally:**\n\n```shell\ntldr --render path/to/page.md\n```\n\n**Print tldr manpage paths as a colon-separated list (see the [Manpage Integration](#Manpage-Integration) section):**\n\n```shell\ntldr --manpath\n```\n\n**Display usage information:**\n\n```shell\ntldr --help\n```\n\n\n### Setting languages\n\n[As specified by the tldr-pages client specification][client-spec-language],\ntldr-pages from other languages can be displayed by this client\n(falling back to English if the page doesn\'t exist for that language).\n\nTo do so, set any of the environment variables `$LANG`, `$LANGUAGE`, or `$TLDR_LANGUAGE` to the two-letter language code\nfor your language (e.g. `export LANGUAGE=es`),\nor set the `--language` option when running `tldr` (e.g. `tldr <COMMAND> --language es`).\n\n\n### Setting platforms\n\nBy default, tldr-pages will be displayed based on your current platform.\nTo directly specify what platform\'s page to use, use the `--platform` flag.\n\nFor example, to display the macOS version of the `top` command\'s tldr-page, run `tldr top --platform macos`.\nThis is the default behavior on macOS,\nbut `--platform macos` is required to show the macOS version of this page on other platforms.\n\n\n## Manpage Integration\n\nThe command `man` can be set up to fall back to displaying tldr-pages if no manpages are found.\n\nTo do so,\nadd the provided line to your shell\'s startup script (e.g. `~/.bash_profile`, `~/.zshenv`, `~/.config/fish/config.fish`)\nto add this behavior to `man`:\n\n### Bash and Zsh\n\n```shell\nexport MANPATH="$MANPATH:$(tldr --manpath)"\n```\n\n### Fish\n\n```shell\nset -gxa MANPATH (tldr --manpath)\n```\n\n[tldr-pages]: https://github.com/tldr-pages/tldr\n[client-spec]: https://github.com/tldr-pages/tldr/blob/main/CLIENT-SPECIFICATION.md\n[client-spec-language]: https://github.com/tldr-pages/tldr/blob/main/CLIENT-SPECIFICATION.md#language\n[xdg]: https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html\n',
    'author': 'Ethan Kinnear',
    'author_email': 'contact@superatomic.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://tldr-man.superatomic.dev/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
